==> create-web-components-project-0.0.2.tgz: package/src/index.js <==
#! /usr/bin/env node

import fs, { existsSync } from 'fs';
import { readdir, readFile, writeFile, mkdir } from 'node:fs/promises';
import path from 'path';
import inquirer from 'inquirer';
import process from 'process';
import { fileURLToPath } from 'url';
import { execa } from 'execa';
import Listr from 'listr';
import PackageJson from '@npmcli/package-json';

const __filename = fileURLToPath(import.meta.url);
const __dirname = path.dirname(__filename);
const CURR_DIR = process.cwd();

function kebabToPascal(str){
  return str
    .split('-')
    .map(item => item.charAt(0).toUpperCase() + item.slice(1).toLowerCase())
    .join("");
}

function kebabToCamel(str) {
  return str
    .replace(/-./g, x=>x[1].toUpperCase());
}

/**
 * @typedef CopyFilesOptionsType
 * @type {object}
 * @property {array.<ReplaceFilenamesType>} replaceFilenames - an array of objects, update filename with find and replace
 * @property {array.<ReplaceValuesType>} replaceValues -  an array of objects, find and replace values within the document
 */

/**
 * @type ReplaceFilenamesType
 * @type {object[]}
 * @property {string} filename - filename value to look up
 * @property {string} replacement - replacement filename value
 */

/**
 * @type ReplaceValuesType
 * @type {object[]}
 * @property {string} find - filename value to look up
 * @property {string} replacement - replacement filename value
 */

/**
 * @param {string} readPath
 * @param {string} writePath
 * @param {CopyFilesOptionsType} options - object with support for `replaceFilenames` and `replaceValues` arrays of objects for file output manipulation
 */
async function copyFiles(readPath, writePath, options) {
  try {
    const items = await readdir(readPath);

    for (let item of items) {
      const itemPath = path.join(readPath, item);
      const isDir = fs.statSync(itemPath).isDirectory();

      if (isDir) {
        const dirPath = path.join(writePath, item);
        if (!existsSync(dirPath)) {
          await mkdir(dirPath);
        }
        copyFiles(itemPath, dirPath, options);
      } else {
        let buffer = await readFile(itemPath, 'utf-8');

        if (options?.replaceValues) {
          for (const r of options.replaceValues) {
            buffer = buffer.replaceAll(r.find, r.replacement);
          }
        }

        if (options?.replaceFilenames) {
          for (const f of options?.replaceFilenames) {
            if (item.includes(f.filename)) {
              item = item.replace(f.filename, f.replacement);
            }
          }
        }
        await writeFile(path.join(writePath, item), buffer, { encoding: 'utf-8' });
      }
    }
  } catch (err) {
    console.error(err);
  }
}

async function initGit() {
  const result = await execa('git', ['init'], {
    cwd: CURR_DIR,
  });
  if (result.failed) {
    return Promise.reject(new Error('Failed to initialize Git'));
  }
  return;
}

async function npmInstall(path) {
  const result = await execa('npm', ['install'], {
    cwd: path,
  });
  if (result.failed) {
    return Promise.reject(new Error('Failed to NPM install'));
  }
  return;
}

async function npmBuild(path) {
  const result = await execa('npm', ['run', 'build'], {
    cwd: path,
  });
  if (result.failed) {
    return Promise.reject(new Error('Failed to run NPM \'build\' script'));
  }
  return;
}

async function configPackageJsonFiles(monorepoDir, compRel) {
  const pkgJson = await PackageJson.load(monorepoDir);
  pkgJson.update({
    workspaces: [
      ...pkgJson.content.workspaces,
      (options?.component && options?.output) ? `${options.output}/*` : '',
    ],
    wireit: {
      ...pkgJson.content.wireit,
      build: {
        dependencies: [
          ...pkgJson.content.wireit?.build?.dependencies,
          `${compRel}:build`
        ]
      },
      test: {
        dependencies: [
          ...pkgJson.content.wireit?.test?.dependencies,
          `${compRel}:test`
        ]
      }
    }
  });
  await pkgJson.save();
}

async function buildMonorepo() {
  const filepath = path.join(__dirname, 'templates', 'monorepo');
  const monorepoDir = path.resolve(CURR_DIR, options?.output ?? '');
  const answers = await inquirer.prompt([
    {
      type: 'string',
      name: 'monorepoName',
      message: 'What is the kebab case name of your monorepo project? \'@zywave/*\' is auto prefixed:   \n',
      validate: function (input) {
        // todo: test this regex if matches npm package name rules
        if (/^[a-z0-9\-]{1,50}$/gm.test(input)) return true;
        else return 'Edit project name above, only lowercase letters, numbers and hyphens allowed';
      },
      transformer: (input) => `@zywave/${input}`
    },
    {
      type: 'confirm',
      name: 'initGit',
      message: 'Would you like to add Git to your project?   \n',
      default: true
    }
  ]);
  const tasks = new Listr([
    {
      title: 'Copying project files',
      task: () => copyFiles(filepath, monorepoDir,
        {
          replaceValues: [{ find: 'placeholder-monorepo-name', replacement: answers.monorepoName }]
        }
      ),
    },
    {
      title: 'Initializing Git',
      task: () => initGit(),
      enabled: () => answers.initGit,
    },
  ]);

  await tasks.run();
  console.log('Let\'s setup your first web component...');
  await buildComponent();
  console.log('%s Monorepo project ready', 'DONE');
}

async function buildComponent() {
  const filepath = path.join(__dirname, 'templates', 'component');
  const answers = await inquirer.prompt([
    {
      type: 'string',
      name: 'componentName',
      message: 'What is the kebab case name of your component? Strongly recommend following this custom element naming format \'zywave-tribename-componentname\':   \n',
      validate: function (input) {
        if (/^[a-z][.0-9_a-z]*-[\-.0-9_a-z]*$/gm.test(input)) return true;
        else return 'Edit component name above... Only lowercase letters, numbers and hyphens allowed. A hyphenated compound name is required per custom elements rules, i.e. \'zywave-name\'';
      }
    }
  ]);

  let compDir;
  let monorepoDir;

  if (options?.component && options?.output) {
    monorepoDir = path.resolve(CURR_DIR);
    compDir = path.resolve(monorepoDir, options.output, answers.componentName);
  } else {
    monorepoDir = path.resolve(CURR_DIR, options?.output ?? '');
    compDir = path.resolve(monorepoDir, 'packages', answers.componentName);
  }

  const compRelFromMonorepo = `./${path.relative(monorepoDir, compDir).replace(/[\\/]+/g, '/')}`;

  const tasks = new Listr([
    {
      title: 'Copying project files',
      task: () => {
        mkdir(compDir);
        copyFiles(filepath, compDir,
          {
            replaceFilenames: [{ filename: '_element', replacement: answers.componentName }],
            replaceValues: [
              { find: 'placeholder-component-name', replacement: answers.componentName },
              { find: 'PlaceholderComponentClassName', replacement: kebabToPascal(answers.componentName) }
            ]
          }
        )
      },
    },
    {
      title: 'Configure monorepo package.json',
      task: async () => {
        await configPackageJsonFiles(monorepoDir, compRelFromMonorepo);
      },
    },
    {
      title: 'Installing NPM dependencies',
      task: async () => {
        await npmInstall(monorepoDir);
      },
    },
    {
      title: 'Building monorepo',
      task: async () => {
        await npmBuild(monorepoDir);
      },
    }
  ]);

  await tasks.run();
  console.log('%s Component files setup', 'DONE');
}

const args = process.argv.slice(2);

/**
 * @type {object}
 * @property? {boolean} project - when present creates a component, without flag CLI creates monorepo project
 * @property? {boolean} component - when present creates a project monorepo, without flag CLI creates monorepo project
 * @property? {string} output - string path to where new component will be added to, path must exist
 */
let options = {};

for (const arg of args) {
  const argAndVal = arg.split('=');
  const key = argAndVal[0].split('--')[1] ?? argAndVal;
  // const key = kebabToCamel(argAndVal[0].split('--')[1]);
  const val = key === 'output'
    ? path.normalize(argAndVal[1]) 
    : argAndVal[1];
  options = {
    ...options,
    [key]: val ?? true,
  }
}

options.component ? await buildComponent() : await buildMonorepo();

==> create-web-components-project-0.0.4.tgz: package/src/index.js <==
#! /usr/bin/env node

import fs, { existsSync } from 'fs';
import { readdir, readFile, writeFile, mkdir } from 'node:fs/promises';
import path from 'path';
import inquirer from 'inquirer';
import process from 'process';
import { fileURLToPath } from 'url';
import { execa } from 'execa';
import Listr from 'listr';
import PackageJson from '@npmcli/package-json';

const __filename = fileURLToPath(import.meta.url);
const __dirname = path.dirname(__filename);
const CURR_DIR = process.cwd();

function kebabToPascal(str){
  return str
    .split('-')
    .map(item => item.charAt(0).toUpperCase() + item.slice(1).toLowerCase())
    .join("");
}

function kebabToCamel(str) {
  return str
    .replace(/-./g, x=>x[1].toUpperCase());
}

/**
 * @typedef CopyFilesOptionsType
 * @type {object}
 * @property {array.<ReplaceFilenamesType>} replaceFilenames - an array of objects, update filename with find and replace
 * @property {array.<ReplaceValuesType>} replaceValues -  an array of objects, find and replace values within the document
 */

/**
 * @type ReplaceFilenamesType
 * @type {object[]}
 * @property {string} filename - filename value to look up
 * @property {string} replacement - replacement filename value
 */

/**
 * @type ReplaceValuesType
 * @type {object[]}
 * @property {string} find - filename value to look up
 * @property {string} replacement - replacement filename value
 */

/**
 * @param {string} readPath
 * @param {string} writePath
 * @param {CopyFilesOptionsType} options - object with support for `replaceFilenames` and `replaceValues` arrays of objects for file output manipulation
 */
async function copyFiles(readPath, writePath, options) {
  try {
    const items = await readdir(readPath);

    for (let item of items) {
      const itemPath = path.join(readPath, item);
      const isDir = fs.statSync(itemPath).isDirectory();

      if (isDir) {
        const dirPath = path.join(writePath, item);
        if (!existsSync(dirPath)) {
          await mkdir(dirPath);
        }
        copyFiles(itemPath, dirPath, options);
      } else {
        let buffer = await readFile(itemPath, 'utf-8');

        if (options?.replaceValues) {
          for (const r of options.replaceValues) {
            buffer = buffer.replaceAll(r.find, r.replacement);
          }
        }

        if (options?.replaceFilenames) {
          for (const f of options?.replaceFilenames) {
            if (item.includes(f.filename)) {
              item = item.replace(f.filename, f.replacement);
            }
          }
        }
        await writeFile(path.join(writePath, item), buffer, { encoding: 'utf-8' });
      }
    }
  } catch (err) {
    console.error(err);
  }
}

async function initGit() {
  const result = await execa('git', ['init'], {
    cwd: CURR_DIR,
  });
  if (result.failed) {
    return Promise.reject(new Error('Failed to initialize Git'));
  }
  return;
}

async function npmInstall(path) {
  const result = await execa('npm', ['install'], {
    cwd: path,
  });
  if (result.failed) {
    return Promise.reject(new Error('Failed to NPM install'));
  }
  return;
}

async function npmBuild(path) {
  const result = await execa('npm', ['run', 'build'], {
    cwd: path,
  });
  if (result.failed) {
    return Promise.reject(new Error('Failed to run NPM \'build\' script'));
  }
  return;
}

async function configPackageJsonFiles(monorepoDir, compRel) {
  const pkgJson = await PackageJson.load(monorepoDir);
  pkgJson.update({
    workspaces: [
      ...pkgJson.content.workspaces,
      (options?.component && options?.output) ? `${options.output}/*` : '',
    ],
    wireit: {
      ...pkgJson.content.wireit,
      build: {
        dependencies: [
          ...pkgJson.content.wireit?.build?.dependencies,
          `${compRel}:build`
        ]
      },
      test: {
        dependencies: [
          ...pkgJson.content.wireit?.test?.dependencies,
          `${compRel}:test`
        ]
      }
    }
  });
  await pkgJson.save();
}

async function buildMonorepo() {
  const filepath = path.join(__dirname, 'templates', 'monorepo');
  const monorepoDir = path.resolve(CURR_DIR, options?.output ?? '');
  const answers = await inquirer.prompt([
    {
      type: 'string',
      name: 'monorepoName',
      message: 'What is the kebab case name of your monorepo project? \'@zywave/*\' is auto prefixed:   \n',
      validate: function (input) {
        // todo: test this regex if matches npm package name rules
        if (/^[a-z0-9\-]{1,50}$/gm.test(input)) return true;
        else return 'Edit project name above, only lowercase letters, numbers and hyphens allowed';
      },
      transformer: (input) => `@zywave/${input}`
    },
    {
      type: 'confirm',
      name: 'initGit',
      message: 'Would you like to add Git to your project?   \n',
      default: true
    }
  ]);
  const tasks = new Listr([
    {
      title: 'Copying project files',
      task: () => copyFiles(filepath, monorepoDir,
        {
          replaceValues: [{ find: 'placeholder-monorepo-name', replacement: answers.monorepoName }]
        }
      ),
    },
    {
      title: 'Initializing Git',
      task: () => initGit(),
      enabled: () => answers.initGit,
    },
  ]);

  await tasks.run();
  console.log('Let\'s setup your first web component...');
  await buildComponent();
  console.log('%s Monorepo project ready', 'DONE');
}

async function buildComponent() {
  const filepath = path.join(__dirname, 'templates', 'component');
  const answers = await inquirer.prompt([
    {
      type: 'string',
      name: 'componentName',
      message: 'What is the kebab case name of your component? Strongly recommend following this custom element naming format \'zywave-tribename-componentname\':   \n',
      validate: function (input) {
        if (/^[a-z][.0-9_a-z]*-[\-.0-9_a-z]*$/gm.test(input)) return true;
        else return 'Edit component name above... Only lowercase letters, numbers and hyphens allowed. A hyphenated compound name is required per custom elements rules, i.e. \'zywave-name\'';
      }
    }
  ]);

  let compDir;
  let monorepoDir;

  if (options?.component && options?.output) {
    monorepoDir = path.resolve(CURR_DIR);
    compDir = path.resolve(monorepoDir, options.output, answers.componentName);
  } else {
    monorepoDir = path.resolve(CURR_DIR, options?.output ?? '');
    const packagesDir = path.resolve(monorepoDir, 'packages')
    await mkdir(packagesDir);
    compDir = path.resolve(packagesDir, answers.componentName);
  }

  const compRelFromMonorepo = `./${path.relative(monorepoDir, compDir).replace(/[\\/]+/g, '/')}`;

  const tasks = new Listr([
    {
      title: 'Copying project files',
      task: () => {
        mkdir(compDir);
        copyFiles(filepath, compDir,
          {
            replaceFilenames: [{ filename: '_element', replacement: answers.componentName }],
            replaceValues: [
              { find: 'placeholder-component-name', replacement: answers.componentName },
              { find: 'PlaceholderComponentClassName', replacement: kebabToPascal(answers.componentName) }
            ]
          }
        )
      },
    },
    {
      title: 'Configure monorepo package.json',
      task: async () => {
        await configPackageJsonFiles(monorepoDir, compRelFromMonorepo);
      },
    },
    {
      title: 'Installing NPM dependencies',
      task: async () => {
        await npmInstall(monorepoDir);
      },
    },
    {
      title: 'Building monorepo',
      task: async () => {
        await npmBuild(monorepoDir);
      },
    }
  ]);

  await tasks.run();
  console.log('%s Component files setup', 'DONE');
}

const args = process.argv.slice(2);

/**
 * @type {object}
 * @property? {boolean} project - when present creates a component, without flag CLI creates monorepo project
 * @property? {boolean} component - when present creates a project monorepo, without flag CLI creates monorepo project
 * @property? {string} output - string path to where new component will be added to, path must exist
 */
let options = {};

for (const arg of args) {
  const argAndVal = arg.split('=');
  const key = argAndVal[0].split('--')[1] ?? argAndVal;
  // const key = kebabToCamel(argAndVal[0].split('--')[1]);
  const val = key === 'output'
    ? path.normalize(argAndVal[1]) 
    : argAndVal[1];
  options = {
    ...options,
    [key]: val ?? true,
  }
}

options.component ? await buildComponent() : await buildMonorepo();
